/*
Ben Tribelhorn
6-26-06

Helper classes and math functions for dealing with vision objects.
*/

const double pi = 3.14159265358979;

struct info
{
    double majoraxisangle;
    double majoraxislength;
    double minoraxislength;
};

// note that this changes the input arrays!
// also the inputs need to be doubles....
struct info getInfo(int *x, int *y, int len)
{
  double mypi = pi;
  double meanx = 0.0, meany = 0.0;

  // find the mean of all data points (x[i],y[i])
  for (int i=0 ; i<len ; ++i)
  {
    meanx += x[i];
    meany += y[i];
  }
  meanx /= len;
  meany /= len;

  // subtract off the mean from each data point
  for (int i=0 ; i<len ; ++i)
  {
    x[i] -= meanx;
    y[i] -= meany;
  }

  // find the covariance matrix
  // [ A  B ] == data * data^T
  // [ B  C ]
  double A = 0.0, B=0.0, C=0.0;
  for (int i=0 ; i<len ; ++i)
  {
    A += x[i]*x[i];
    B += x[i]*y[i];
    C += y[i]*y[i];
  }
  A /= len;
  B /= len;
  C /= len;

  // find the two eigenvalues (lambdas)
  double a = 1.0;
  double b = -(A+C);
  double c = A*C-B*B;
  double lambda1 = (-b + sqrt(b*b - 4*a*c)) / (2*a);
  double lambda2 = (-b - sqrt(b*b - 4*a*c)) / (2*a);
  if (lambda1 < lambda2)   // swap the two values
  {  
    double tmp = lambda1;
    lambda1 = lambda2;
    lambda2 = tmp;
  }

  // now we need to find the major axis
  // the equations to solve for x and y are
  //   Ax + By = lambda x      (A-la)x = -By
  //   Bx + Cy = lambda y  or  Bx = (la-C)y
  // this is just atan2...
  double ae1, ae2; // eigenvector angles in radians
  double xe1, ye1, xe2, ye2; // unit eigenvectors 1 and 2, x and y
  ae1 = atan2(lambda1-A,B);
  // make sure we're between -pi/2 and pi/2
  if (ae1 < -mypi/2.0) ae1 += mypi;
  if (ae1 > mypi/2.0) ae1 -= mypi;
  ae2 = ae1 + mypi/2.0;  // between 0 and pi
  xe1 = cos(ae1); ye1 = sin(ae1);
  xe2 = cos(ae2); ye2 = sin(ae2);

  // how do we find the axes' lengths?
  // the eigenvalues are proportional to those lengths, but
  // we really just want to take the length of the projection
  // of each data point onto the appropriately angled axis
  // and then subtract max - min:
  double max1 = 0.0, min1 = 0.0;
  double max2 = 0.0, min2 = 0.0;
  double proj;
  for (int i=0 ; i<len ; ++i)
  {
    proj = x[i]*xe1 + y[i]*ye1;
    if (proj > max1) max1 = proj;
    if (proj < min1) min1 = proj;
    proj = x[i]*xe2 + y[i]*ye2;
    if (proj > max2) max2 = proj;
    if (proj < min2) min2 = proj;
  }

  // we should be all set... ?
  struct info I;
  I.majoraxisangle = ae1;
  I.majoraxislength = max1-min1;
  I.minoraxislength = max2-min2;
  return I; // so small that pass-by-value is probably
            // faster than allocating memory on the heap
            // and then worrying about deleting it
}

struct Component
{
	int pixelCount;
	int avgX, avgY;
	int objColor;
	struct info axes;
	double roundness;
} components[100];

void initComponent(struct Component *c, int pCount, int x, int y, int color, struct info a)
{
	c->pixelCount = pCount;
	c->avgX = x;
	c->avgY = y;
	c->objColor = color;
	c->axes = a;
	c->roundness = 4.0 * pCount / (pi * a.majoraxislength * a.majoraxislength);
}

struct connectedComponent
{
	int x[307300];
	int y[307300];
	int nextEL;
} cc;

void ccClear(struct connectedComponent *obj)
{	
	//for(int i = 0; i< obj->nextEL; i++)
	//	obj->x[i] = obj->y[i] = 0;
	obj->nextEL = 0;
}

void ccInit(struct connectedComponent *obj)
{
	const int SIZE = 307300;
	obj->nextEL = 0;
	for(int i = 0; i< SIZE; i++)
		obj->x[i] = obj->y[i] = 0;
}

inline void ccAddPixel(struct connectedComponent *obj, int X, int Y)
{
	obj->x[obj->nextEL] = X;
	obj->y[obj->nextEL] = Y;
	(obj->nextEL)++;
}

void ccPrint(struct connectedComponent *obj)
{
	for(int i = 0; i< obj->nextEL; i++)
	{
		printf("pixel # %d: (%d, %d)\n", i, obj->x[i], obj->y[i]);
	}
}


